"""
Anonymise all personal data in the database
"""
from django.core.management import BaseCommand
from anonymizable_model import settings as app_settings


def get_anonymized_models():
    from django.apps import apps
    anonymized_models = []
    for app in apps.get_app_configs():
        for model in app.get_models():
            if hasattr(model, "db_label_prefix") and hasattr(model, "anonymizable_fields"):
                anonymized_models.append(model)
    return anonymized_models


class Command(BaseCommand):
    help = "Anonymises all personal data in the database"

    def add_arguments(self, parser):
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            default=True,
            help="Tells Django to NOT prompt the user for input of any kind.",
        )

    def handle(self, *args, **options):
        if not app_settings.DJANGO_ANONYMIZE_CAN_ANONYMIZE_DATABASE:
            raise ValueError("Database anonymisation is not enabled")
        interactive = options["interactive"]

        if interactive:  # pragma: no cover
            confirm = input(
                """Warning!
                You have requested that all personal information in the database is anonymised.
                This will IRREVERSIBLY OVERWRITE all personal data currently in the database.
                Are you sure you want to do this?
                Type 'yes' to continue, or 'no' to cancel: """
            )
        else:
            confirm = "yes"

        if confirm == "yes":
            msg = ""
            for model in get_anonymized_models():
                def anonymize_field(obj, field):
                    display_method = getattr(obj, "display_anonymized_"+ field)
                    setattr(obj, field, display_method())
                    obj.save(update_fields=[field])

                [[anonymize_field(obj, field) for field in model.anonymizable_fields] for obj in model.objects.all()]
                msg += "{} model anonymized. \n".format(model._meta.verbose_name)
            self.stdout.write(msg)

        else:  # pragma: no cover
            self.stdout.write("Anonymisation cancelled.")


# -*- coding: utf-8 -*-
import copy
import logging

from django.utils.translation import gettext as _

logger = logging.getLogger(__name__)


class AnonymizableAdminMixin:
    DISPLAY_ANONYMIZE_PREFIX = "display_anonymized_"

    def _display_anonymized_string(self, obj):
        if obj and obj.id:
            return _(f"Anonymized Object {obj.id}")
        return _("Anonymized Object")

    def is_anonymizable(self):
        if hasattr(self.model, "anonymizable_fields"):
            return True
        else:
            logger.error(
                _(
                    f"anonymizable_fields not set in {self.model}, "
                    f"add anonymizable decorator or register your model with anonymizable.register method"
                )
            )
            raise Exception(
                _(
                    f"anonymizable_fields not set in {self.model}, "
                    f"add anonymizable decorator or register your model with anonymizable.register method"
                )
            )

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = super().get_readonly_fields(request, obj)
        readonly_fields = copy.deepcopy(readonly_fields)
        if self.is_anonymizable() and not request.user.has_perm("example.can_view_anonymized_fields"):
            readonly_fields = self.add_anonymizable_fields_to_readonly_fields(readonly_fields, self.model.anonymizable_fields)
        return readonly_fields

    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj)
        fieldsets = copy.deepcopy(fieldsets)
        if self.is_anonymizable() and not request.user.has_perm("example.can_view_anonymized_fields"):
            fieldsets = self.substitute_anonymizable_fields_from_fieldset(fieldsets, self.model.anonymizable_fields)
        return fieldsets

    def get_list_display(self, request):
        list_display = super().get_list_display(request)
        list_display = copy.deepcopy(list_display)
        if self.is_anonymizable() and not request.user.has_perm("example.can_view_anonymized_fields"):
            list_display = self.add_anonymizable_fields_to_list_display(list_display, self.model.anonymizable_fields)
        return list_display

    def render_change_form(self, request, context, add=False, change=False, form_url='', obj=None):
        if obj and obj.id and not request.user.has_perm("example.can_view_anonymized_fields"):
            context.update({
                "subtitle": _(f"Anonymized Object {obj.id}"),
            })
        return super().render_change_form(request, context, add, change, form_url, obj)

    @classmethod
    def add_anonymizable_fields_to_list_display(cls, list_display, anonymizable_fields):
        for anonymizable_field in anonymizable_fields:
            list_display = list(list_display)
            if "__str__" in list_display:
                list_display[list_display.index("__str__")] = "_display_anonymized_string"
            if anonymizable_field in list_display:
                list_display[list_display.index(anonymizable_field)] = cls.DISPLAY_ANONYMIZE_PREFIX + anonymizable_field
        return tuple(list_display)

    @classmethod
    def add_anonymizable_fields_to_readonly_fields(cls, readonly_fields, anonymizable_fields):
        for anonymizable_field in anonymizable_fields:
            readonly_fields = list(readonly_fields)
            if cls.DISPLAY_ANONYMIZE_PREFIX + anonymizable_field not in readonly_fields:
                readonly_fields.append(cls.DISPLAY_ANONYMIZE_PREFIX + anonymizable_field)
        return tuple(readonly_fields)

    @classmethod
    def substitute_anonymizable_fields_from_fieldset(cls, fieldsets, anonymizable_fields):
        for fieldset_name, fieldsets_fields in fieldsets:
            fields = list(fieldsets_fields.get("fields", []))
            for idx, item in enumerate(fields):
                if isinstance(item, (tuple, list)):
                    items = list(item)
                    for index, field in enumerate(items):
                        if field in anonymizable_fields:
                            items[index] = cls.DISPLAY_ANONYMIZE_PREFIX + field
                    fields[idx] = tuple(items)
                elif isinstance(item, str):
                    if item in anonymizable_fields:
                        fields[idx] = cls.DISPLAY_ANONYMIZE_PREFIX + item
            fieldsets_fields["fields"] = tuple(fields)
        return fieldsets





from django.utils.translation import gettext as _


def anonymizable(db_label_prefix="pa_", anonymizable_fields=None):

    def display_anonymized_FIELD(self):
        return "..."

    def wrap(model_class):
        model_class.db_label_prefix = db_label_prefix
        model_class.anonymizable_fields = anonymizable_fields
        new_db_table_name = db_label_prefix + model_class._meta.db_table
        model_class._meta.db_table = new_db_table_name
        model_class._meta.permissions = [
            ("can_view_anonymized_fields", "Can view anonymized fields",)
        ]
        model_class._meta.original_attrs.update({
            "db_table": new_db_table_name
        })
        model_class._meta.original_attrs.update({
            "permissions": [("can_view_anonymized_fields", "Can view anonymized fields",)]
        })
        
        # for m2m fields
        for m2m_field in model_class._meta.local_many_to_many:
            if m2m_field:
                if not m2m_field.db_table and db_label_prefix not in m2m_field.name:
                    m2m_field.db_table = f"{model_class._meta.app_label}_{model_class._meta.model_name}_{m2m_field.name}"
                elif m2m_field.db_table:
                    m2m_field.db_table.replace(db_label_prefix, "")

        if "display_anonymized_FIELD" not in model_class.__dict__:
            setattr(
                model_class,
                "display_anonymized_FIELD",
                display_anonymized_FIELD
            )
        for field in anonymizable_fields:
            model_field = model_class._meta.get_field(field)
            new_field_column_name = db_label_prefix + model_field.column
            model_field.db_column = new_field_column_name
            model_field.column = new_field_column_name
            # For Admin visualization
            if 'display_anonymized_%s' % field not in model_class.__dict__:
                def wrapped_display_anonymized_function(self):
                    return self.display_anonymized_FIELD()
                setattr(
                    model_class,
                    'display_anonymized_%s' % field,
                    wrapped_display_anonymized_function,
                )
                method = getattr(model_class, 'display_anonymized_%s' % field)
                method.short_description = _(model_field.verbose_name)
        return model_class
    return wrap



